function [L,D,p] = Mldl( A,tol,thresh )

% tol=1e-3;  thresh=0.01;  [L,D,p] = Mldl( A,tol,thresh );
% creates a sparse positive definite preconditioner M = L*D*L'
% from a symmetric indefinite matrix A using Matlab's ldl function.
% With precon = 1 and MLBL.m, M maybe used as a preconditioner with minres
% for solving indefinite Ax = b.

% Drops elements of A smaller than tol.
% Gets LDL' factorization of A, where D has 1x1 and 2X2 blocks.
% Changes D to |D| by modifying the eigensystem of each block.
%
% On entry:
% A       is a scaled symmetric indefinite matrix (use scaleSID( A )).
% tol     is the tolerance for creating A2 = sparsified A.
% thresh  is the pivoting threshold for [L,D,p] = ldl(A2,thresh,'vector');
%         (thresh is in [0.0.5], default is 0.01).
%
% On exit:
% L       is a sparse lower-triangular matrix.
% D       is a sparse positive-definite block-diagonal matrix.
% p       is a permutation vector such that M = L*D*L' approximates A(p,p).
%         (p is a row vector.)

% Maintainers: Shaked Regev and Michael Saunders, ICME, Stanford University.
% 07 May 2020: First version of sparselbl.m.
% 13 May 2020: First version of Mldl.m with D block-diagonal (not diagonal).

  n       = length(A);
  nnzA    = nnz(A);
  d       = diag(A);      % sparse column vector
  A       = tril(A,-1);   % Lower triangular part of A, excluding diagonal
  [I,J,V] = find(A);
  small   = find(abs(V) <= tol);
  nsmall  = 2*length(small);   % L + L'
  if nsmall>0
     I(small) = [];
     J(small) = [];
     V(small) = [];
     A = sparse(I,J,V,n,n);
  end
  A = A + A' + diag(d);   % exactly symmetric
  fprintf('\n nnz(A) = %9i      n   = %9i'  , nnzA,n)
  fprintf('\n nnz(A2)= %9i', nnz(A))
  fprintf('\n nsmall = %9i      tol = %9.1e', nsmall,tol)

  [L,D,p] = ldl(A,thresh,'vector');
  nsing   = 0;
  ntwo    = 0;
  tolsing = 1e-6;

  i = 1;
  while i<n
     i1 = i+1;
     a  = full(D(i,i));
     b  = full(D(i,i1));
     if b~=0
        ntwo  = ntwo + 1;
        c     = full(D(i1,i1));
        [V,E] = eig([a b; b c]); % E = dense 2-vector
        e     = diag(E);         %   = roots of r^2 - (a+c)r + (ac-b^2) = 0
        e     = abs(e);
        if e(1) < tolsing, e(1) = tolsing;  nsing = nsing+1; end
        if e(2) < tolsing, e(2) = tolsing;  nsing = nsing+1; end
        D(i:i1,i:i1) = V*diag(e)*V';   % sparse 2x2 block = dense 2 x 2
        i = i+2;
     else
        a = abs(a);
        if a < tolsing, a = tolsing;  nsing = nsing+1; end
        D(i,i) = a;
        i = i1;
     end
  end
  if i==n
     a = abs(full(D(n,n)));
     if a < tolsing, a = tolsing;  nsing = nsing+1; end
     D(n,n) = a;
  end
  fprintf('\n nnz(L) = %9i     ntwo = %9i     from ldl(A2)', nnz(L),ntwo)
  fprintf('\n nsing  = %9i  tolsing = %9.1e\n', nsing, tolsing)

end % function Mldl
