function M = Mminres3( A )

%        M = Mminres3( A );
%
% A = sparse, symmetric, maybe indefinite matrix.  Should be scaled by sscale.m
%
% 26 Apr 2020: First version of Mminres3, derived from Mminres2.
%              Calls minres3 to do 3 iterations of Minres for each column of M.
% 28 Apr 2020: Cols of M sorted if necessary to keep nnz(m) <= lfill.
%              Allow for unit vectors in A causing r=0.
% 07 May 2020: Output M is exactly symmetric.

  n      = length(A);
  lfill  = min(n,ceil(nnz(A)/n)+2);   % Avg nonzeros per column
  M      = spalloc(n,n,n*lfill);      % n columns
  temp   = spalloc(n,1,2*lfill);      % 1 column
  z0     = 0;
  tol    = 1e-3;    % Relative tolerance for ignoring small nonzeros in M.
  nsmall = 0;
  nsort  = 0;

  for j=1:n
    % nz   = nnz(A(:,j));
    % some initial guess
    m      = temp;
    m(j)   = 1;
    r      = m - A(:,j);   % add -A*m if not using 0 as initial guess.

    if (norm(r,inf) > tol)
       % Mminres3 uses 3 iterations of minres on A*x = r, then mj = mj + x.

       x = minres3( A,r );
       m = m + x;

       % Sparsify m.
       newm    = false;      % Track if m is altered.
       [I,J,V] = find(m);
       lenV    = length(V);
       absV    = abs(V);
       reltol  = tol*norm(V,inf);
       small   = find(absV <= reltol);
       lsmall  = length(small);
       if lsmall>0
          nsmall   = nsmall + lsmall;
          newm     = true;
          I(small) = [];
          J(small) = [];
          V(small) = [];
       end

       % Keep the lfill largest nonzeros.
       if length(V) > lfill
          nsort = nsort + 1;
          newm     = true;
          [W,P] = sort(abs(V),'descend');
          P     = P(1:lfill);
          I     = I(P);
          J     = J(P);
          V     = V(P);
       end

       if newm, m = sparse(I,J,V,n,1); end

       prnt = mod(j,10000)==1;
       if prnt
          fprintf('\n j=%g lfill=%5i  lenV=%7i lsmall=%8i nnz(m)=%8i nnz(M)=%8i', ...
                      j,   lfill,     lenV,    lsmall,    nnz(m),    nnz(M));
       end
    end % if norm(r,inf) > tol

    if m(j)==0
       z0     = z0+1;
    elseif m(j)>0
       M(:,j) =  m;
    else
       M(:,j) = -m;
    end
  end % loop on j

  M = (M+M')*0.5;
  fprintf('\n')
  fprintf('\n nnz(A) = %9i   nnz(M) = %9i', nnz(A),nnz(M))
  fprintf('\n nsmall = %9i      tol = %9.1e', nsmall, tol)
  fprintf('\n Sorted = %9i   Fractn = %9g', nsort, nsort/n)
  fprintf('\n Number of zero diagonal entries = %d', z0)
  fprintf('\n Mminres3: M symmetrized')

  % Sparsify symmetrized M.
  d = diag(M);           % sparse column vector
  M = tril(M,-1);
  [I,J,V] = find(M);
  small   = find(abs(V) <= tol);
  nsmall  = length(small);
  if nsmall>0
     I(small) = [];
     J(small) = [];
     V(small) = [];
     M        = sparse(I,J,V,n,n);
  end
  M = M + M' + diag(d);  % M is exactly symmetric
  fprintf('\n nsmall = %9i      tol = %9.1e\n', nsmall, tol)

end % function Mminres3
