function [ x, istop, itn, rnorm, Arnorm, Anorm, Acond, xnorm, betacheck ] = ...
           minres( x0, A, b, M, precon, shift, show, check, itnlim, rtol, tolMPD )

%        [ x, istop, itn, rnorm, Arnorm, Anorm, Acond, xnorm, betacheck ] = ...
%          minres( x0, A, b, M, precon, shift, show, check, itnlim, rtol, tolMPD );
%
% minres solves the n x n system of linear equations Ax = b
% or the n x n least squares problem           min ||Ax - b||_2^2,
% where A is a symmetric matrix (possibly indefinite or singular) of order n
% and b is a given vector of dimension n = length(b).
% If shift is nonzero, "A" in the above problem definitions becomes A - shift*I.
%
% INPUT:
%
% x0     is an estimate of x, or [] if an estimate is not available.
%        minres computes r0 = b - A*x0, solves A dx = r0, and updates x := x0 + dx.
%        This guards against loss of low-order digits of dx if x0 is a good approximation
%        (r0 and dx are small).
%
% A      may be a dense or sparse matrix (preferably sparse!)
%        or a function handle such that y = A(x) returns the product y = A*x
%        for any given n-vector x.  (This is true even if shift is nonzero.)
%
% b      is the right-hand side n-vector.
%        Normally b will be dense, though Matlab might do the right thing
%        in some cases if b is sparse.
%
% M      is an optional preconditioner (or []).  Ideally, M should be SPD.
%        Depending on precon, M may be a dense or sparse matrix that approximates
%        A or inv(A) (really A - shift*I or inv(A - shift*I)),
%        or it may be a function handle such that y = M(x) solves My = x
%        or computes y = M*x.
%
% precon = 0 or M = [] implies no preconditioning.  The input M is not used.
%
% precon = 1 implies M = CC' ~= A and minres implicitly solves the system
%                               PAP'xbar = Pb,
%        i.e.                  Abar xbar = bbar,
%        where                         P = inv(C),
%                                   Abar = PAP',
%                                   bbar = Pb,
%        and returns the solution      x = P'xbar.
%        The associated residual is rbar = bbar - Abar xbar
%                                        = P(b - A*x).
%        Note that minres asks for systems M*x = y to be solved via y = M\x.
%        This will be efficient if you set M = decomposition(M1,'chol'); say,
%        meaning that you know M1 ~= A is SPD and chol(M1) is practical.
%
%        If A2 ~= A is SPD and it is feasible to factorize [L,D,P,S] = LDL(A2,thresh)
%        (say), it is possible to set M = decomposition(A2,'ldl'); and input M.
%        It will use the default options for ldl(A2) such as thresh = 0.01.
%        (We don't know how to change those.)
%        If A2 is indefinite, we would like to change D (with 1x1 and 2x2 blocks)
%        to positive definite D2 = |D| (using abs(eigenvalues of D)),
%        but we don't know how to replace D by D2 in the output of "decomposition".
%
% precon = 2 implies M = P'P ~= inv(A) and (again) minres implicitly solves the system
%                               PAP'xbar = Pb,
%        i.e.                  Abar xbar = bbar,
%        where                      Abar = PAP',
%                                   bbar = Pb.
%        and returns the solution      x = P'xbar.
%        The associated residual is rbar = bbar - Abar xbar
%                                        = P(b - Ax).
%        Note that minres asks for products y = M*x.  This is efficient if M is a
%        sparse matrix.
%
% shift  If shift ~= 0, minres treats "A" as (A - shift*I).
%        If shift is an eigenvalue of A, x should approximate an eigenvector of A.
%
% show   = 0 suppresses output.
%        > 0 prints early iterations, then occasional intermediate iterations,
%            then final iterations for which tolerances are nearly satisfied.
%
% check  = 0 suppresses preliminary checks on A and M.
%        > 0 requests an initial check on the symmetry of A and M
%            and the positive definiteness of M.
%        The test
%
% itnlim is a limit on the number of minres iterations.
%        Theoretically, itnlim = n should be large enough, but in practice,
%        minres may need many more or many fewer than n iterations to converge.
%        If O(n) iterations are needed, a better preconditioner M is needed.
%
% rtol   is a tolerance on certain residual norms.
%        Let r = b - Ax, bnorm = ||b||, rnorm = ||r||, Arnorm = ||A*r||, xnorm = ||x||.
%        If there is no preconditioning, minres stops if rnorm or Arnorm is sufficiently
%        small (allowing for bnorm and xnorm and an estimate of ||A||).
%        With preconditioning, rnorm = ||r||M or ||Arnorm ||_M should be sufficiently small.
%        Set rtol in [1e-12,1e-3] (say).
%
% tolMPD is a tolerance on certain inner products that should be positive if M is SPD.
%        Set tolMPD in [1e-6,1e-2] (say).  See output betacheck.
%
% OUTPUT:
%
% x      is the final estimate of the required solution
%        after k iterations, where k is return in itn.
%
% istop  is a value from [-1:10] to indicate the reason for termination.
%        The reason is summarized in msg[istop+2] below.
%
% itn    gives the final value of k (the iteration number).
%
% rnorm  estimates norm(r_k)  (or norm(rbar_k) if M exists).
%
% Arnorm estimates norm(Ar_{k-1}) or norm(Abar rbar_{k-1}) if M exists.
%        NOTE THAT Arnorm LAGS AN ITERATION BEHIND rnorm.
%
% xnorm  is the final norm(x).
%
% betacheck indicates whether M seems to be indefinite.  It is normalized
%        and may be treated as an absolute number.
%        If precon==1, betacheck = y'inv(M)y / y'y for some y.
%        If precon==2, betacheck =      y'My / y'y for some y. 
%        If itn==0, y = b.
%        Output istop==10 means betacheck < tolMPD.
%        With tolMPD in [1e-6,1e-2] (say), this indicates that
%        M is not safely positive definite.  Setting
%           gamma = facMPD*(tolMPD - betacheck);   M = M + gamma*speye(n);
%        with facMPD in [10,100] (say) will make M more SPD.

% Code author:  Michael Saunders, SOL and ICME, Stanford University
% Contributors: Chris Paige, School of Computer Science, McGill University
%               Sou-Cheng Choi, ICME, Stanford University
%               Shaked Regev,   ICME, Stanford University 
%
% 02 Sep 2003: Date of Fortran 77 version, based on 
%              C. C. Paige and M. A. Saunders (1975),
%              Solution of sparse indefinite systems of linear equations,
%              SIAM J. Numer. Anal. 12(4), pp. 617-629.
%
% 02 Sep 2003: ||Ar|| now estimated as Arnorm.
% 17 Oct 2003: f77 version converted to MATLAB.
% 03 Apr 2005: A must be a matrix or a function handle.
% 10 May 2009: Parameter list shortened.
%              Documentation updated following suggestions from
%              Jeffery Kline <jeffery.kline@gmail.com>
%              (author of new Python versions of minres, symmlq, lsqr).
% 06 Jul 2009: Michael Chen <mc462@cornell.edu> reports divide by zero
%              when beta = 0 (in this case it was beta_2 = 0).
%              Realized that the istop values were out of sync.
%              They should be right now.
% 02 Sep 2011: David Fong reports error in Acond when alpha1=0.
%              gmax and gmin should be initialized before itn 1.
% 02 Sep 2011: ynorm = norm(x) is now computed directly instead of
%              being updated (incorrectly).
% 19 Mar 2020: x0 and precon are new input parameters.
% 24 Mar 2020: Input tolMPD and output betacheck cater for unintentionally indefinite M.
% 03 Apr 2020: If b is sparse, some quantities stay sparse for a while.
%              With v and y sparse, alfa = v'*y;       gives a sparse scalar(!).
%              We need              alfa = full(v'*y);
% 20 Apr 2020: The cheaply updated rnorm and the test for rnorm being small enough
%              (for given rtol) used to apply to the preconditoned system (not ideal).
%              Now when rnorm seems small enough, the true r = b - A*x and rnorm
%              are computed.  If necessary, rtol is reduced to rtol/10 up to 5 times.
%              Hence: the input rtol now applies to Ax=b, but the rnorm is updated cheaply
%              on most iterations.
% 26 Apr 2020: If b is sparse, treat v and w as sparse until Matlab changes them to dense.
%              Trying to allow for applications that do only 2 or 3 iterations.


% Known bugs:
%  1. As Jeff Kline pointed out, Arnorm = ||A r_{k-1}|| lags behind
%     rnorm = ||r_k||.  On singular systems, this means that a good
%     least-squares solution exists before Arnorm is small enough
%     to recognize it.  The solution x_{k-1} gets updated to x_k
%     (possibly a very large solution) before Arnorm shuts things
%     down the next iteration.  It would be better to keep x_{k-1}.
%------------------------------------------------------------------

% Initialize                               

  msg = [' beta2 = 0.  If M = I, b and x are eigenvectors '   % -1
         ' beta1 = 0.  The exact solution is  x = 0       '   %  0
         ' A solution to Ax = b was found, given rtol     '   %  1
         ' A least-squares solution was found, given rtol '   %  2
         ' Reasonable accuracy achieved, given eps        '   %  3
         ' x has converged to an eigenvector              '   %  4
         ' Acond has exceeded 0.1/eps                     '   %  5
         ' The iteration limit was reached                '   %  6
         ' A  does not define a symmetric matrix          '   %  7
         ' M  does not define a symmetric matrix          '   %  8
         ' M  is not sufficiently SPD                     '   %  9
         ' rtol reduced max times on preconditioned system'   % 10
        ];
 

  n      = length(b);
  r0     = b;
  bnorm  = norm(b);
  r0norm = bnorm;
  x0norm = 0;

  if length(x0)==n
    y      = minresxxxA( A,x0 ) - shift*x0;
    r0     = b - y;
    r0norm = norm(r0);
    x0norm = norm(x0);
  end

  if isempty(M), precon = 0; end
  nnzr0    = nnz(r0);
  sparser0 = nnzr0/n <= 0.2;

  if show
    fprintf('\n minres.m   SOL, Stanford University   Version of 07 May 2020')
    fprintf('\n Solution of symmetric Ax = b or (A-shift*I)x = b')
    fprintf('\n n       =%9g     shift   =%22.14e',      n, shift )
    fprintf('\n itnlim  =%9g     rtol    =%9.1e'  , itnlim, rtol  )
    fprintf('\n precon  =%9g     tolMPD  =%9.1e'  , precon, tolMPD)
    fprintf('\n norm(x0)=%9.1e     norm(b) =%9.1e', x0norm, bnorm )
    fprintf('\n nnz(r0) =%9i     norm(r0)=%9.1e'  , nnzr0 , r0norm)
  end

  istop  = 0;   itn    = 0;
  rnorm  = 0;   Arnorm = 0;
  Anorm  = 0;   Acond  = 0;
  xnorm  = 0;   betacheck = 0;
  dxnorm = 0;
  done   = false;

  %---------------------------------------------------------------------
  % Decode A.
  %---------------------------------------------------------------------
  if show
    if isa(A,'double')         % A is an explicit matrix A.
      if issparse(A)
        fprintf('\n A is an explicit sparse matrix')
        fprintf('    nnz(A) =%10i', nnz(A))
      else
        fprintf('\n A is an explicit dense matrix' )
      end
    elseif isa(A,'function_handle')
      disp(['A is defined by function_handle ' func2str(A)])
    else
      error('minres','A must be a matrix or a function handle')
    end

    if issparse(b)
      fprintf('\n b is sparse')
    else
      fprintf('\n b is dense')
    end
  end

  %---------------------------------------------------------------------
  % Decode M.
  %---------------------------------------------------------------------
  if precon>0
    if isa(M,'double')       % M is an explicit matrix M.
      if issparse(M)
        fprintf('\n M is an explicit sparse matrix')
        fprintf('    nnz(M) =%10i', nnz(M))
      else
        fprintf('\n M is an explicit dense matrix' )
      end
    elseif isa(M,'function_handle')
      fprintf('\n M is defined by function_handle %s', func2str(M))
    elseif isa(M,'decomposition')
      fprintf('\n M is a decomposition for use with precon=1')
    else
      error('minres','M must be a matrix or decomposition or function handle')
    end
  end

  %------------------------------------------------------------------
  % Set up y and v for the first Lanczos vector v1.
  % y  =  beta1 P' v1,  where  P = C**(-1).
  % v is really P' v1.
  %------------------------------------------------------------------
  % We solve A*x = r0 and later set x = x0 + x.
  if sparser0
    x = sparse(n,1);
  else
    x = zeros(n,1);   
  end
  y   = r0;
  r1  = r0;
  if precon>0, y = minresxxxM( precon,M,r0 ); end
  beta1     = r0'*y;              beta1     = full(beta1);
  betacheck = beta1 / (r0'*r0);   betacheck = full(betacheck);
                                                
  % Test for an indefinite preconditioner.
  % If r0 = 0 exactly, stop with x = x0.

  if beta1==0
    istop = 0;    show = true;   done = true;
  elseif precon>0 && betacheck < tolMPD
    istop = 9;    show = true;   done = true;
  end

  if ~done      % beta1> 0
    beta1  = sqrt(beta1);       % Normalize y to get v1 later.

    % See if M is symmetric.

    if check & precon
      r2     = minresxxxM( precon,M,y );
      s      = y' *y;      s = full(s);
      t      = r1'*r2;     t = full(t);
      z      = abs(s-t);
      if z > 0
        z   = z / (abs(s)+abs(t));
        tol = eps^(1/3);
        if z > tol, istop = 8;  show = true;  done = true; end
      end
    end

    % See if A is symmetric.

    if check
      w    = minresxxxA( A,y );
      r2   = minresxxxA( A,w );
      s    = w'*w;        s = full(s);
      t    = y'*r2;       t = full(t);
      z    = abs(s-t);
      if z > 0
        z   = z / (abs(s)+abs(t));
        tol = eps^(1/3);
        if z > tol, istop = 7;  done  = true;  show = true;  end
      end
    end
  end % ~done

  %------------------------------------------------------------------
  % Initialize other quantities.
  % ------------------------------------------------------------------
  oldb   = 0;       beta   = beta1;   dbar = 0;       epsln = 0;
  qrnorm = beta1;   phibar = beta1;   rhs1 = beta1;
  rhs2   = 0;       tnorm2 = 0;       gmax = 0;       gmin  = realmax;
  cs     = -1;      sn     = 0;
  numrtol= 1;     % Counts initial rtol and later reduced values if any.
  rtol0  = rtol;  % Save original rtol.
  r2     = r0;
  if sparser0
    w  = sparse(n,1);
    w2 = sparse(n,1);
  else
    w  = zeros(n,1);
    w2 = zeros(n,1);
  end

  if show
    fprintf('\n\n    Itn     x(1)     Compatible    LS       norm(A)  cond(A)')
  % fprintf(' gbar/|A|\n')   %%%%%% Check gbar
    x1 = 0;   if x0norm > 0, x1 = x0(1); end
    fprintf('\n %6g %12.5e', itn,full(x1));
  end

  %---------------------------------------------------------------------
  % Main iteration loop.
  % --------------------------------------------------------------------
  if ~done            
    while itn < itnlim                  % k = itn = 1 first time through
      itn    = itn+1;

      %-----------------------------------------------------------------
      % Obtain quantities for the next Lanczos vector vk+1, k = 1, 2,...
      % The general iteration is similar to the case k = 1 with v0 = 0:
      %
      %   p1      = Operator * v1  -  beta1 * v0,
      %   alpha1  = v1'p1,
      %   q2      = p2  -  alpha1 * v1,
      %   beta2^2 = q2'q2,
      %   v2      = (1/beta2) q2.
      %
      % Again, y = betak P vk,  where  P = C**(-1).
      % .... more description needed.
      %-----------------------------------------------------------------
      s = 1/beta;                 % Normalize previous vector (in y).
      v = s*y;                    % v = vk if P = I

      y = minresxxxA( A,v ) - shift*v;
      if itn >= 2, y = y - (beta/oldb)*r1; end

      alfa   = full(v'*y);        % alphak
      y      = (- alfa/beta)*r2 + y;
      r1     = r2;
      r2     = y;
      if precon>0,  y = minresxxxM( precon,M,r2 );  end
      oldb   = beta;              % oldb = betak
      beta   = full(r2'*y);       % beta = betak+1^2
      betacheck = beta / full(r2'*r2);

      if precon>0 && betacheck < tolMPD   % M is not sufficiently SPD
        istop = 9;   show = true;   done = true;
        x1   = x(1);
        if x0norm > 0, x1 = x0(1) + x1; end
        fprintf('\n %6g %12.5e  betacheck=%10.3e', itn,full(x1),betacheck);
        break;
      end

      beta   = sqrt(beta);
      tnorm2 = tnorm2 + alfa^2 + oldb^2 + beta^2;

      if itn==1                   % Initialize a few things.
        if beta/beta1 <= 10*eps   % beta2 = 0 or ~ 0.
          istop = -1;             % Terminate later.
        end
      end

      % Apply previous rotation Qk-1 to get
      %   [deltak epslnk+1] = [cs  sn][dbark    0   ]
      %   [gbar k dbar k+1]   [sn -cs][alfak betak+1].

      oldeps = epsln;
      delta  = cs*dbar + sn*alfa; % delta1 = 0         deltak
      gbar   = sn*dbar - cs*alfa; % gbar 1 = alfa1     gbar k
      epsln  =           sn*beta; % epsln2 = 0         epslnk+1
      dbar   =         - cs*beta; % dbar 2 = beta2     dbar k+1
      root   = norm([gbar dbar]);
      Arnorm = phibar*root;       % ||Ar{k-1}||

      % Compute the next plane rotation Qk

      gamma  = norm([gbar beta]); % gammak
      gamma  = max([gamma eps]);
      cs     = gbar/gamma;        % ck
      sn     = beta/gamma;        % sk
      phi    = cs*phibar ;        % phik
      phibar = sn*phibar ;        % phibark+1

      % Update x.

      denom  = 1/gamma;
      w1     = w2;
      w2     = w;
      w      = (v - oldeps*w1 - delta*w2)*denom;
      x      = x + phi*w;

      % Go round again.

      gmax   = max([gmax gamma]);
      gmin   = min([gmin gamma]);
      z      = rhs1/gamma;
      rhs1   = rhs2 - delta*z;
      rhs2   =      - epsln*z;

      % Estimate various norms (for preconditioned system).
      % Arnorm for previous iteration is estimated above.

      Anorm  = sqrt( tnorm2 );
      dxnorm = norm(x);   % Really norm(dx), excludes x0
      rnorm  = phibar;

      % Estimate cond(A).
      % In this version we look at the diagonals of  R  in the
      % factorization of the lower Hessenberg matrix,  Q * H = R,
      % where H is the tridiagonal matrix from Lanczos with one
      % extra row, beta(k+1) e_k^T.

      Acond  = gmax/gmin;
      if istop ~= 0, break; end   % Exit while itn < itnlim

      %================================================================
      % Check stopping criteria.
      % In rare cases, istop is already -1 from above (Abar = const*I)
      % or istop = 7 (A not symmetric) or 8 (M not symmetric).
      %================================================================
    % epsa   = (Anorm + beta1)*eps;
      epsx   = (Anorm*dxnorm + beta1)*eps;
      epsr   = (Anorm*dxnorm + beta1)*rtol;
      test1  = rnorm/(Anorm*dxnorm + bnorm); % ||r||/(||A|| ||x|| + ||b||) for precond system
    % test2  = root / Anorm;                 % ||Ar{k-1}||/(||A|| ||r_{k-1}||)    "   "   " 
      test2  = Arnorm / (Anorm*(rnorm + eps));
      t1 = 1 + test1;      % These tests work if rtol < eps
      t2 = 1 + test2;
      if t2    <= 1      , istop = 2; end
      if t1    <= 1      , istop = 1; end
     
      if itn   >= itnlim , istop = 6; end
      if Acond >= 0.1/eps, istop = 5; end
      if epsx  >= beta1  , istop = 3; end
    % if rnorm <= epsx   , istop = 2; end
    % if rnorm <= epsr   , istop = 1; end
      if test2 <= rtol   , istop = 2; end
      if test1 <= rtol   , istop = 1; end
 
      if precon>0 & istop>0 & istop<=5
         % Preconditioned system thinks it satisfied rtol.
         % See if Ax = b is satisfied to rtol0.
         if x0norm==0
            xnorm  = dxnorm;
            rnormk = norm(b - minresxxxA( A,x ) - shift*x);
         else
            xt     = x0 + x;
            xnorm  = norm(xt);
            rnormk = norm(b - minresxxxA( A,xt ) - shift*xt);
         end
         epsr   = (Anorm*xnorm + bnorm)*rtol0;
         if rnormk <= epsr
            istop = 1;
            fprintf('\n rnormk small enough: %8.1e    epsr =%8.1e', rnormk, epsr)
         elseif numrtol<5 & rtol>eps   % Reduce rtol
            numrtol = numrtol + 1;
            rtol    = rtol/10;
            fprintf('\n%7g %2i: rtol reduced to%8.1e', itn,numrtol, rtol)
            fprintf('    rnormk =%8.1e    epsr =%8.1e', rnormk, epsr)
            istop = 0;
         else
            istop = 10;
         end
      end

      % See if it is time to print something.

      prnt   = false;
      if n      <= 40       , prnt = true; end
      if itn    <= 10       , prnt = true; end
      if itn    >= itnlim-10, prnt = true; end
      if mod(itn,1000)==0   , prnt = true; end
      if rnorm  <=1.001*epsx, prnt = true; end
      if rnorm  <=1.001*epsr, prnt = true; end
      if Acond  >= 1e-2/eps , prnt = true; end
      if istop  ~=  0       , prnt = true; end

      if show & prnt
        if mod(itn,10)==0, disp(' '); end
        x1   = x(1);
        if x0norm > 0, x1 = x0(1) + x1; end
        str1 = sprintf('%6g %12.5e %10.3e', itn,full(x1),test1);
        str2 = sprintf(' %10.3e',           test2);
        str3 = sprintf(' %8.1e %8.1e',      Anorm,Acond);
        str4 = sprintf(' %8.1e',            gbar/Anorm);
        str  = [str1 str2 str3];
      % str  = [str1 str2 str3 str4];
        fprintf('\n %s', str)

        debug = 0;
        if debug   % Print true Arnorm.
                   % This works only if no preconditioning.
          vv = b - minresxxxA(A,x)  + shift*x;    % vv = b - (A - shift*I)*x
          ww =     minresxxxA(A,vv) - shift*vv;   % ww = (A - shift*I)*vv = "Ar"
          trueArnorm = norm(ww);
          fprintf('\n Arnorm = %12.4e   True ||Ar|| = %12.4e', Arnorm,trueArnorm)
        end
      end % show & prnt

      if istop ~= 0, break; end

    end % while itn < itnlim (main loop)
  end % if ~done early

% Display final status.

  xnorm  = dxnorm;
  if x0norm>0, x = x0 + x;  xnorm = norm(x); end
  y      = minresxxxA( A,x ) - shift*x;
  r1     = b - y;
  rnorm  = norm(r1);
  y      = minresxxxA( A,r1) - shift*r1;
  Arnorm = norm(y);

  if show
    fprintf('\n\n istop   =%3g               itn   =%6g', istop,itn  )
    if precon>0
       fprintf('\n Abarnorm=%12.4e   Abarcond =%12.4e', Anorm,Acond)
    else
       fprintf('\n Anorm   =%12.4e      Acond =%12.4e', Anorm,Acond)
    end
    fprintf('\n rnorm   =%12.4e      xnorm =%12.4e', rnorm,xnorm)
    fprintf('\n Arnorm  =%12.4e\n', Arnorm)
    disp(msg(istop+2,:))
  end
end
%-----------------------------------------------------------------------
% End function minres
%-----------------------------------------------------------------------


function y = minresxxxA( A,x )

% sets   y = A*x for a matrix A defined by input parameter A.
%
% 10 May 2009: A is now a matrix or a function handle.

  if isa(A,'double')
    y = A*x;
  else
    y = A(x);
  end
end
%-----------------------------------------------------------------------
% End private function minresxxxA
%-----------------------------------------------------------------------


function y = minresxxxM( precon,M,x )

% If precon = 1, solve My = x, where M is the preconditioner.
% If precon = 2, return y = Mx.
%
% 10 May 2009: M is now a matrix or a function handle.
% 19 Mar 2020: Allow M to approximate A (precon=1) or inv(A) (precon=2).

  if isa(M,'double')
     switch(precon)
        case 1
           y = M\x;
        case 2
           y = M*x;
     end
  elseif isa(M,'decomposition')
     y = M\x;
  else   % M is a function for solving M*y = x
     y = M(x);
  end
end
%-----------------------------------------------------------------------
% End private function minresxxxM
%-----------------------------------------------------------------------
